package Elchemea;
#$Id: Elchemea.pm,v 1.80 2018/11/29 08:56:09 sqko Exp $

#################################################################
#
#
#   Package Elchemea
#
# The Elchemea.pm module contains a number of utility functions for the
# elchemea system. the functions are described below:
# 
#
# -- \begin{longtable}{ l p{8cm} }
# -- debug(opt $level)  & Turns debug on and off. Returns the debug level if called without an argument.\\
# -- set_multiplex_info(@) & Writes the arguments to the multiplex file. One argument at each line .\\
# -- get_multiplex_array() & Returns an array with the content of the multiplex file used to store the multiplexer setting.\\
# -- get_multiplex_info() & Returns the first line of the multiplexfile used to store the multiplexer setting.\\
# -- get_external_polarisation_function($) & Returns a function pointer to the run_ext_cmd function. Based on the input string it sets the program to be executed on function call (sets the $ext_pol_cmd variable).\\
# -- get_last_message() & Returns the content of the last_message variable.\\
# -- set_last_message($) & Sets the last message variable to the specified argument.\\
# -- append_last_message($) & Appends the specified string to the last_message variable.\\
# -- unset_message() & Deletes the content of the last_message variable .\\
# -- is_locked() & Returns true if a measurement is running.\\
# -- is_locked_device() & Returns true if a device is processing commands (this will be the case for some seconds after a measurement is stopped before cleanup is complete).\\
# -- lock_measure() & Marks a measurment as running and locks the device.\\
# -- unlock_measure() & Removes the seamafore marking that a measuremnt is running.\\
# -- unlock_device() & Removes the seamafore marking the device as in use.\\
# -- stop_measure() & Stops the current measurement.\\
# -- get_config_value($$) & Returns the configuration value for the specified section and key (See the get_cv function in ElchemeaConfig.pm for details).\\
# -- stop_program() & Stops any running program. Note does NOT stop any chronoamperometry with no finite runtime. Stops normal chronoamperometry though.\\
# -- stop_chronoforever() & Stops any chronoamperometry with no finite runtime.\\
# -- get_dir_list() & Returns a list of files in the specified directory which matches the specified regular expression.\\
# -- mail($$$) & Sends an email to the specified address. Arguments: address, message, subject.\\
# -- EFAlog($) & Appends the specified string to the error file (includes stack backtrace).\\
# -- get_time($) & Returns the time in format YYYY MM DD HH MM SS. Output is in the form of an array. Input is either 'now' or the offset (check 'man date' for details).\\
# -- CGI_client(@) & returns the result of the specified command and optional arguments to the CGI-server.\\
# --\end{longtable}
#
# User administration functions. Note that some of these functions are only used if the user authentification is based on a local user list.
#
# -- \begin{longtable}{ l p{10cm} }
# -- add_user($) & Function which adds a user by adding a line to the userlist (passwd) file. The function also creates the user directry and copies the configuration files to the relevant positions.\\
# -- user_exists($) & Checks if the user exists. Returns 1 if true 0 otherwise .\\
# -- get_pwd($) & Returns the password hash for the specified user.\\
# -- set_pwd($$) & Sets the password hash for the specified user to the specified value. Arguments: username, passwdhash.\\
# -- set_proglog($) & Writes the specified string to the 'current command file.\\
# -- get_priglog() & Returns the content of the 'current_command' file.\\
# -- get_remaining_time() & Returns the remaining time in seconds. Looks in the proglog file for index 2.\\
# -- format_time($) & returns a text string describing the time supplied (Expects time in seconds). Note that the returned string will only be an approximate value (returns '1 hour 5 minutes' for 1 hour, 5 minutes and 16 seconds for instance).\\
# --\end{longtable}
#
#################################################################

$ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin';
use strict;
use warnings;
use Fcntl qw(:DEFAULT :flock);
use POSIX;
use Socket;
use ElchemeaHeader qw(:DEFAULT);
use ElchemeaDevice;
use ElchemeaConfig;
use SemaforeFile;
use SocketClient;
use RemoteExec;
use File::Copy;
use Mail::Send;

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = $ElchemeaHeader::VERSION;
    our $COPYRIGHT = $ElchemeaHeader::COPYRIGHT;
    our $DESIGNERS = $ElchemeaHeader::DESIGNERS;
    our $IMPLEMENTERS = $ElchemeaHeader::IMPLEMENTERS;
    @EXPORT_OK = qw(EFAlog serial_client socket_client CGI_client get_config_value lock_measure unlock_measure lock_device unlock_device);
    %EXPORT_TAGS = (
        'standard'=>[qw/EFAlog get_config_value lock_measure unlock_measure serial_client/]
		    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

#### Global file locations ####
my $configfile = $ElchemeaHeader::configfile;
my $configlock = $ElchemeaHeader::configlock;
my $command_lock = $homedir.'current_command.lock';
my $command_file = SemaforeFile->new($homedir.'current_command.txt',$command_lock);
my $passwdlock = $homedir.'pwd.lock';
my $passwdfile = SemaforeFile->new($homedir.'users.txt',$passwdlock);
my $plotfile = $htmldir.'png/current.png';
my $multiplexfile = SemaforeFile->new($homedir.'multiplex.nr');
#### End global file locations ####

#### Private variables #####
my $global = ElchemeaConfig->new($configfile);
my $last_message = undef;
my $debug = 0;

return 1;

sub debug
{
    if (@_)
    {
	$debug = shift;
	$multiplexfile->debug($debug);
	$command_file->debug($debug);
	$passwdfile->debug($debug);
	$global->debug($debug);
    }
    return $debug;
}

sub set_multiplex_info(@)
{
    ##################################################
    ### set_multiplex_info(@)                      ###
    ### Writes the arguments to the multiplex file ###
    ### one argument at each line                  ###
    ##################################################
    $multiplexfile->writeline(@_);
}

sub get_multiplex_array()
{
    ####################################################
    #### get_multiplex_info()                        ###
    #### returns an array with the content of the    ###
    #### multiplex file used to sotre the            ###
    #### multiplexer setting                         ###
    ####################################################
    return () unless ($multiplexfile->exist);
    return $multiplexfile->readlines();
}

sub get_multiplex_info()
{
    ####################################################
    #### get_multiplex_info()                        ###
    #### returns the first line of the multiplexfile ###
    #### used to sotre the multiplexer setting       ###
    ####################################################
    my @info = &get_multiplex_array();
    return $info[0] if (@info);
    return '';
}

sub get_external_polarisation_function($)
{
    ############################################################
    ### get_external_polarisation_function($)                ###
    ### Returns a function pointer to the run_ext_cmd        ###
    ### function. Based on the input string it sets the      ###
    ### program to be executed on function call              ###
    ### (sets the $ext_pol_cmd variable)                     ###
    ############################################################
    my $input = shift;
    return undef unless ($input);
    my $retval = '';
    if ($input && $input =~ /([\w\d\s\t\.\-\+]+)/)
    {
	$input = $1;
	my ($file,@args) = split(/[\s\t]+/,$input);
	my $cmdfile = $user_exec_dir.$file;
	if (-e $cmdfile && -x $cmdfile)
	{
	    my $ext_pol_cmd = $user_exec_dir.$file.' '.join(' ',@args);
	    $retval = RemoteExec->new($ext_pol_cmd);
	}
	else
	{
	    $retval = 'File does not exist';
	}
    }
    else
    {
	$retval = 'No match';
    }
    return $retval;
}

sub get_last_message()
{
    ##################################
    ### get_last_message()         ###
    ### returns the content of the ###
    ### last_message variable      ###
    ##################################
    my $val = $last_message;
    $val = '' unless (defined($val));
    return $val;
}

sub set_last_message($)
{
    $last_message = shift;
}

sub append_last_message($)
{
    if (defined($last_message))
    {
	$last_message .= shift;
    }
    else
    {
	$last_message = shift;
    }
}


sub unset_message()
{
    $last_message = undef;
}

sub is_locked()
{
    return -e $measurelock;
}

sub is_locked_device()
{
    return -e $devicelock;
}

sub lock_measure()
{
    return if (-e $measurelock);
    open MYOUT,'>',$measurelock or die "Could not open file $measurelock for writing $!";
    close MYOUT;
    chmod 0666,$measurelock;
    &lock_device();
    return $measurelock;
}

sub unlock_measure()
{
    &stop_measure();
    &unlock_device();
}

sub lock_device()
{
    open MYOUT,'>',$devicelock or die "Could not open file $devicelock for writing $!";
    close MYOUT;
    chmod 0666,$devicelock;
}

sub unlock_device()
{
    unlink $devicelock if (-e $devicelock);
}

sub stop_measure()
{    
    unlink $measurelock if (-e $measurelock);
}

sub get_config_value($$)
{
    return $global->get_config_value(@_);
}

sub stop_program()
{
    #################################################
    ### stop_program()                            ###
    ### stops any running program, note does NOT  ###
    ### stop any chronoamperometry with no finite ###
    ### runtime                                   ###
    ### Stops normal chronoamperometry though     ###
    #################################################
    unlink $programlock if -e $programlock;
}

sub stop_chronoforever()
{
    ##################################################
    ### stop chronoforever()                       ###
    ### Stops any chronoamperometry with no finite ###
    ### runtime                                    ###
    ##################################################
    unlink $chrono_runforever_semafore if -e $chrono_runforever_semafore;
}

sub get_dir_list($@)
{
    my $dir = shift;
    return undef unless (-d $dir);
    my $regexp = '.*';
    $regexp = shift if (@_);
    opendir DIR,$dir or die EFAlog("Could not open dir $dir for reading!\n");
    my @list = grep {!/^\.{1,2}$/ && -f $dir.'/'.$_} readdir(DIR);
    closedir DIR;
    my @retdir = ();
    foreach my $f (@list)
    {
        push @retdir,$f if ($f =~ /$regexp/);
    }
    return @retdir;
}


sub mail($$$)
{
    my $email = shift;
    my $message = shift;
    my $subject = shift;
    ### To do: Insert true email handling routine ##
    my $msg = Mail::Send->new();
    $msg->to($email);
    $msg->subject($subject);
    my $fh = $msg->open;
    print $fh $message;
    $fh->close or die "Could not send message to $email $!\n";
}

sub EFAlog($)
### Function which writes the input string to te errorlog #####
{
    my $string = shift;
    chomp $string;
    #### Gets file locations ####
    my $file = $homedir.'error.txt';
    my $lock = $homedir.'errorlog.lock';
    ### Locks file access for other processes ###
    sysopen(FH, $lock, O_WRONLY | O_CREAT) or die "Error: can't open file: $lock $!\n";
    flock(FH, LOCK_EX);
    ### Opens the errorlog file #####
    open(ELOG, '>>', $file) or die "Error: can't open logfile: $file $!\n";
    my $time = localtime;
    ### Prints the errormessage ####
    print ELOG $time, " : ";
    print ELOG $string;
    #######################################
    ## Getting complete stack backtrace ###
    ## for anything else than warnings  ###
    #######################################
    my @trace = ();
    my $i = 0;
    while(my ($package,$filename,$line,$subroutine) = caller($i++))
    {
        unshift @trace,"'$filename' line $line called $subroutine";
    }
    print ELOG ' #### Stack backtrace: ';
	print ELOG join(' -> ',@trace);
    print ELOG "\n";
    close(ELOG);
    flock(FH, LOCK_UN);
    close(FH);
    return 1;
}

sub get_time($)
### Function which returns the time in format YYYY MM DD HH MM SS      ###
### Output is in the form of an array                                  ###
### Input is either 'now' or the offset (check 'man date' for details) ###
{
    my $tsh = shift;
    my $ts = 0;
    my $u = 'min';
    if ($tsh =~ /([\-\+\d]+)(\w+)/)
    {
        $ts = $1+0;
        $u = $2;
    }
    $ts = $ts * 60 if ($u =~ /min/);
    $ts = $ts * 60 * 60 if ($u =~ /hou/);
    my @t = localtime(time + $ts);
    $t[5] += 1900;
    $t[4]++;
    my $time = sprintf("%04d %02d %02d %02d %02d %02d",$t[5],$t[4],$t[3],$t[2],$t[1],$t[0]);
    @t = split(/\s+/,$time);
    return @t;
}

####### Section 3: User administration section #############

sub add_user
#### Function which adds a user by adding a line to the userlist (passwd)   ####
#### file. The function also creates the user                               ####
#### Dirwectry and copies the configuration files to the relevant positions ####
{
    my $tuser = shift;
    $tuser =~ /([\d\w]+)/;
    my $user = $1;
    if (length($user) < 4)
    {
	EFAlog("Username '$user' too short for new user (min 4 letters)!");
        return 0;
    }
    if (user_exist($user))
    {
	EFAlog("User '$user' already exists!");
        return 0;
    }
    my $pwd = $user.'123';
    $pwd = shift if (@_);
    my $pass = crypt($pwd,$user);
    $passwdfile->append($user.':'.$pass);
    my $dir = $userdir.$user;
    mkdir $dir,0777 unless -d $dir;
    my $file = $userdir.$user.'/test.nr';
    unless (-e $file)
    {
	### Skip if exist already, which may be the case when restoring a system ###
	open MYOUT,'>',$file or die "Could not open file $file for writing $!";
	print MYOUT "0\n";
	close MYOUT;
    }
    $file = $userdir.$user.'/program.pl';
    unless (-e $file)
    {
	### Skip if exist already, which may be the case when restoring a system ###
	open MYOUT,'>',$file or die "Could not open file $file for writing $!";
	close MYOUT;
	chmod 0666,$file;
    }
    my $source = $homedir.'config';
    ## Use site specified default confuguration if it exists ##
    $source = $homedir.'site_config' if (-e $homedir.'site_config');
    my $target = $userdir.$user.'/config';
    ### Skip if exist already, which may be the case when restoring a system ###
    copy ($source,$target) unless (-e $target);
    return 1;
}

sub user_exist($)
### Checks if the user exists          ###
### Returns 1 if true 0 otherwise      ###
{
    my $user = shift;
    my $res = 0;
    $passwdfile->lock_sh;
    $passwdfile->open_read;
    while(my $str = $passwdfile->readline)
    {
	chomp $str;
	next unless ($str =~ /^$user:/);
	$res = 1;
	last;
    }
    $passwdfile->close;
    $passwdfile->unlock;
    return $res;
}

sub get_pwd($)
### Returns the password hash for the specified user ###
{
    my $user = shift;
    my $res;
    $passwdfile->lock_sh;
    $passwdfile->open_read;
    while(my $str = $passwdfile->readline)
    {
	chomp $str;
	next unless ($str =~ /^$user:([\w\.\-]+)/);
	$res = $1;
	last;
    }
    $passwdfile->close;
    $passwdfile->unlock;
    return $res;
}

sub set_pwd($$)
{
    my $user = shift;
    my $pass = shift;
    my $p = crypt($pass,$user);
    my $res = "User $user not found!\n";
    my @list = $passwdfile->readlines();
    foreach my $line (@list)
    {
	next unless ($line =~ /^$user:/);
	$line = $user.':'.$p;
	$res = "Changed password for $user!\n";
	last;
    }
    $passwdfile->writeline(@list);
    return $res;
}

sub set_proglog($@)
#### Function which writes a string to the 'current command file' ####
{
    my $endtime = shift;
    my $time = time;
    my @arr = @_;
    foreach (@arr)
    {
	$_ = ' ' unless ($_);
    }
    my @lines = ($time,$endtime,@arr);
    $command_file->writeline(@lines);
}


sub get_proglog()
{
    return $command_file->readlines();
}

sub get_remaining_time()
{
    ### Returns the remaining time in seconds ###
    ### looks in the proglog file for index 2 ###
    my @l = get_proglog();
    my $time = time;
    my $tleft = 0;
    $tleft = $l[1] - $time if ($l[1]);
    return $tleft;
}

sub format_time($)
{
    ### returns a text string describing the time supplied ###
    ### (Expects time in seconds)                          ###
    my $time = shift;
    $time = sprintf("%i",$time);
    return '0 seconds' if ($time < 0);
    return $time.' seconds' if ($time < 60);
    if ($time < 300)
    {
	my $min = POSIX::floor($time/60);
	my $sec = $time - $min*60;
	return "$min minutes $sec seconds";
    }
    if ($time < 3600)
    {
	my $min = POSIX::floor($time/60);
	return "$min minutes";
    }
    if ($time < 3*3600 )
    {
	my $hour = POSIX::floor($time/3600);
	my $min = POSIX::floor(($time - $hour*3600)/60);
	return "$hour hours $min minutes";
    }
    if ($time < 86400)
    {
	my $hour = POSIX::floor($time/3600);
	return "$hour hours";
    }
    if ($time < 200000)
    {
	my $day = POSIX::floor($time/86400);
	my $hour = POSIX::floor(($time - $day*86400)/3600);
	return "$day days $hour hours";
    }
    my $day = POSIX::floor($time/86400);
    return "$day days";
}


##### socket communication ###



sub CGI_client(@)
{
    my $remote  =  get_config_value('CGI-server','host');
    my $port    =  get_config_value('CGI-server','port');
    return socket_client($remote,$port,@_);
}
