package Solartron1255;
#$Id: Solartron1255.pm,v 1.62 2017/10/26 14:18:02 sofc Exp $

############################################################
#
#    Package Solartron1255
#
# This module implements the FRA class for a Solartron
# --{\textregistered} 
# 1252 and 1255/1255B frequency response analysers.
#
# The class Solartron1255 handles both versions of the 1255 and the 
# Solartron1252 class handles the 1252.
#
# To get a device instance of one of the classes call the constructors:
#
# $fra = Solartron1255->new($gpib_address);
#
# $fra = Solartron1252->new($gpib_address);
#
############################################################


use strict;
use warnings;
use FDEV;
our @ISA = qw(FRA);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1255A';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V'};
    $self->{'_minf'} = 10E-6;
    $self->{'_maxf'} = 20E+6;
    $self->{'_minpV'} = -40.95;
    $self->{'_maxpV'} = 40.95;
    $self->{'_minpA'} = undef;
    $self->{'_maxpA'} = undef;
    $self->{'_maxampA'} = undef;
    $self->{'_maxampV'} = 3;
    $self->{'_source_modes'} = { 'V1' => '1,0' , 'V2' => '2,0' , 'V1/V2' => '1,2' , 'V2/V1' => '2,1'};
    $self->{'_warningcodes'} = {'-1' => 'Low limit', '00' => ' ', '+1' => 'High limit'};
    $self->{'_errorcodes'} = {
	'00' => ' ',
	'01' => 'Unknown command',
	'02' => 'Argument mismatch',
	'03' => 'Argument out of range',
	'04' => 'Floating point format error',
	'05' => 'Ilegal request for value',
	'06' => 'Invalid function',
	'07' => 'Number out of range',
	'08' => 'Invalid symbol',
	'09' => 'Amplitude illegal for f > 10MHz',
	'31' => 'Generator killed',
	'32' => 'Generator overload',
	'34' => 'Generator restart',
        'XX' => 'No data recieved',
	'80' => ' ', 
	'81' => 'Input Overload',
	'82' => 'Auto integration failed',
	'83' => 'Input Overload and Autointegration failed',
	'84' => 'Generator killed',
	'85' => 'Input Overload and Monitor fail',
	'86' => 'Autointegration failed and Monitor fail',
	'87' => 'Input overload, Monitor fail and Autointegration failed',
	'88' => 'Error 88 Not defined',
	'89' => 'Group dealy parameter not set'
    };
    $self->{'_overhead'} = 1.1;
    $self->{'_minintegration'} = 1;
    $self->{'_setuptime'} = 6;
    $self->{'_delay'} = 0.1;  ## Default delay of 100 ms ##
    $self->{'_setup_tags'} = 'mode,input_mode_V1,input_mode_V2,polarisation,polarisation_offset,source_mode,outer_mode_V1,outer_mode_V2,input_coupling_V1,input_coupling_V2,measuring_range_V1,measuring_range_V2,overhead,amplitude,delay';
    $self->{'_default'} = {
        'mode' => 'Volt',
        'input_mode_V1' => 'Differential',
        'input_mode_V2' => 'Differential',
        'source_mode' => 'V2/V1',
        'outer_mode_V1' => 'Grounded',
        'outer_mode_V2' => 'Grounded',
        'input_coupling_V1' => 'DC',
        'input_coupling_V2' => 'DC',
        'measuring_range_V1' => 'auto',
        'measuring_range_V2' => 'auto',
        'polarisation_offset' => 'OFF',
        'reference_potential' => 'Fixed',
	'delay' => $self->{'_delay'},
	'amplitude' => $self->{'_minampV'},
	'polarisation' => 0,
    };
    $self->{'_desc'} = {
        'mode' => 'Control mode',
        'input_mode_V1' => 'Input mode for V1 input, either differential or single ended',
        'input_mode_V2' => 'Input mode for V2 input, either differential or single ended',
        'source_mode' => 'Determines how the impedance is calculated',
        'outer_mode_V1' => 'Determines if the screen on voltage input 1 is grounded or left floating',
        'outer_mode_V2' => 'Determines if the screen on voltage input 2 is grounded or left floating',
        'input_coupling_V1' => 'Determines if the input coupling for voltage input 1 is direct through or uses a high pass filter',
        'input_coupling_V2' => 'Determines if the input coupling for voltage input 2 is direct through or uses a high pass filter',
        'measuring_range_V1' => 'Input range for voltage input 1, either fixed or autorange',
        'measuring_range_V2' => 'Input range for voltage input 2, either fixed or autorange',
        'polarisation_offset_program' => 'Select if a polarisation offset program (and which one) is to be run prior to impedance scans',
        'reference_potential' => 'Select if the polarisation is to be set versus chassis ground or versus measured potential at OCV',
        'delay' => 'Delay in seconds between each frequency',
        'amplitude' => 'AC amplitude (RMS, V or mA depending on mode)',
        'polarisation' => 'DC polarisation (V or mA depending on mode)',
    };
    $self->{'_values_source_mode'} = ['V1','V2','V1/V2','V2/V1'];
    $self->{'_values_outer_mode_V1'} = ['Floating','Grounded'];
    $self->{'_values_outer_mode_V2'} = ['Floating','Grounded'];
    $self->{'_values_input_coupling_V1'} = ['AC','DC'];
    $self->{'_values_input_coupling_V2'} = ['AC','DC'];
    $self->{'_values_measuring_range_V1'} = ['auto',0.03,0.3,3];
    $self->{'_values_measuring_range_V2'} = ['auto',0.03,0.3,3];
    $self->{'_values_input_mode_V1'} = ['Differential','Single'];
    $self->{'_values_input_mode_V2'} = ['Differential','Single'];
    $self->{'_save_tags'} = ['FR','VA','VB','MS','RA1','RA2','DC1','DC2','IP1','IP2','OU1','OU2','SO','CV','VI'];
    bless $self,$class;
    return $self;
}

sub bulk_load_check(@)
{
    my $self = shift;
    $self->writecmd('CE');
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    $self->writecmd('OP2,1');  # GPIB does all #
    while (@_)
    {
        my $cmd = shift;
        $self->checkcmd($cmd);
    }
    return undef;
}

sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
        $self->writecmd($tag.'?');
        my $r = $self->readcmd();
        if ($r =~ /([\d\.\+\-eE\,]+)/)
        {
            push @res,$tag.' '.$1;
        }
    }
    return @res;
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    $self->writecmd('OP2,1');  # GPIB does all #
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /solartron/i);
    return 0;
}


sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    $self->writecmd('*RST');
    sleep(3);
    $self->writecmd('BP0');
    $self->writecmd('BP0');
    $self->writecmd('CE');
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    $self->writecmd('OP2,1');  # GPIB does all #
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    $self->writecmd('VA 0.000');
    $self->writecmd('VB 0.000');
    return $res;
}

sub reset()
{
    my $self = shift;
    $self->writecmd('BK');
    sleep(1);
    $self->writecmd('BK');
    sleep(1);
}

sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $a = shift;
    $a = $self->{'_minampV'} if ($a < $self->{'_minampV'});
    $a = $self->{'_maxampV'} if ($a > $self->{'_maxampV'});
    $self->checkcmd(sprintf("VA %.4g",$a));
    $self->{'_amplitude'} = $a;
    return $self->{'_amplitude'};
}

sub set_polarisation($$)
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
#    if (ref($ext_pol) eq 'CODE')
#    {
#	$self->writedebug("Executeing function to get polarisation offset");
#        $self->{'_pol_offset'} = $ext_pol->($self);
#        $pol += $self->{'_pol_offset'};
#    }
    if (ref($ext_pol) eq 'RemoteExec')
    {
	$self->writedebug("Executeing function to get polarisation offset");
        $self->{'_pol_offset'} = $ext_pol->exec_id($self);
        $pol += $self->{'_pol_offset'};
    }
    $pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
    $pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'});
    $self->checkcmd(sprintf("VB %.4g",$pol));
    return $pol;
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('ER?');
    my $rep = $self->readcmd();
    if (!$rep || $rep !~ /^00/)
    {
	$self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '$rep'");
	if ($rep = /~(\d+)/)
	{
	    $self->errorlog('Error description: '.$self->errorcode($1));
	}
	$self->writecmd('CE');
    }
}

sub get_error()
{
    my $self = shift;
    $self->writecmd('ER?');
    my $rep = $self->readcmd();
    if ($rep =~ /(\d+)/)
    {
        my $r = $self->errorcode($1);
        $self->writedebug("Error description: $r");
        return $r;
    }
    return '';
}


sub checkcmd($)
{
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    $self->writecmd('ER?');
    my $e = $self->readcmd();
    if (!$e || $e !~ /^00/)
    {
	if ($e = /~(\d+)/)
	{
	    $self->writedebug('Error description: '.$self->errorcode($1));
	}
	$self->writecmd('CE');
	$self->writecmd($cmd);
	$self->checkerror();
    }
}

sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    my $diff1 = 1;
    my $diff2 = 1;
    my $input = 1;
    $diff1 = 0 if ($self->get_setup_value($self->get_default('input_mode_V1'),$setup,'input_mode_V1') =~ /S/);
    $diff2 = 0 if ($self->get_setup_value($self->get_default('input_mode_V2'),$setup,'input_mode_V2') =~ /S/);
    $self->checkcmd("IP1,$diff1");
    $self->checkcmd("IP2,$diff2");
    $self->checkcmd('OT3'); # Fast reply: cr + EOI #
    $self->checkcmd('OP2,1');  # GPIB does all #
    $self->checkcmd('WF0');  ## Sine wave ##
    $self->checkcmd('AU0');  ## Autointegration off ##
    $self->checkcmd('MD0');  ## Normal mode impedance ##
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    $self->checkcmd(sprintf("MS %.4g",$self->{'_delay'})); 
    my $source_mode = $self->get_setup_value($self->get_default('source_mode'),$setup,'source_mode');
    ### Default V2/V1 ###
    my $sm = '2,1';
    $sm = $self->{'_source_modes'}{$source_mode} if exists($self->{'_source_modes'}{$source_mode});
    $self->checkcmd('SO'.$sm);
    $self->checkcmd('CV0');  ## Display a,b ##
    $self->checkcmd('VI0'); ## Display frequency ##
    $self->checkerror();
    ### Screen to ground default on input ###
    my $ou1 = 0;
    $ou1 = 1 if ($self->get_setup_value($self->get_default('outer_mode_V1'),$setup,'outer_mode_V1') =~ /F/i);
    my $ou2 = 0;
    $ou2 = 1 if ($self->get_setup_value($self->get_default('outer_mode_V2'),$setup,'outer_mode_V2') =~ /F/i);
    $self->checkcmd('OU1,'.$ou1);
    $self->checkcmd('OU2,'.$ou2);
    ### DC coupling default ###
    my $ac1 = 0;
    $ac1 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V1'),$setup,'input_coupling_V1') =~ /AC/i);
    my $ac2 = 0;
    $ac2 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V2'),$setup,'input_coupling_V2') =~ /AC/i);
    $self->checkcmd('DC1,'.$ac1);
    $self->checkcmd('DC2,'.$ac2);
    ### Measure range on input ###
    my $ra1 = $self->get_setup_value($self->get_default('measuring_range_V1'),$setup,'measuring_range_V1');
    if ($ra1)
    {
	my $tr = 0; ## Default autorange ##
	my $x = 0;
	foreach (@{$self->{'_values_measuring_range_V1'}})
	{
	    $tr = $x if ($_ eq $ra1);
	    $x++
	}
	$ra1 = $tr;
    }
    my $ra2 = $self->get_setup_value($self->get_default('measuring_range_V2'),$setup,'measuring_range_V2');
    if ($ra2)
    {
	my $tr = 0; ## Default autorange ##
	my $x = 0;
	foreach (@{$self->{'_values_measuring_range_V2'}})
	{
	    $tr = $x if ($_ eq $ra2);
	    $x++
	}
	$ra2 = $tr;
    }
    $self->checkcmd('RA1,'.$ra1);
    $self->checkcmd('RA2,'.$ra2);
    $self->checkcmd('SW0'); ## Sweep off (we do all)
    ### Set polarisation and amplitude ###
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol,$setup);
    $self->amplitude($self->get_setup_value(0,$setup,'amplitude'));
    $self->reset_data;
    return $truepol;
}

sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});    
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    $self->checkcmd('FR '.$fout);
    $self->checkcmd('IS '.$tout);
    $self->writecmd('SI');  ## Run impedance ###
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
	sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->();
        }
        next if (!$testfile || -e $testfile);
        $self->reset();
	push @{$self->{'_data'}},undef;
        return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
	my $readloop = 0;
      INNER:
	while ($readloop < 5)
	{
	    $self->writecmd('DO'); ## write all to gpib ##
	    my $res = $self->readcmd();
	    if (!$res || $res =~ /error/i)
	    {
		$readloop++;
		next INNER;
	    }
	    my @r = split(",",$res);
		# sometimes the reading is an response of the old recording, in that case a new reading of the solartron buffer give the new recording result
		# !note if single frequency measurement is done this will not trap old data reading from the past measurement
		if($r[0] != $fout){
				$self->writecmd('DO'); ## write all to gpib ##
				$res = $self->readcmd();	## perhaps check for error again
				@r = split(",",$res);
		}
	    my $result = sprintf("%f %f %f",$r[0],$r[1]*$f,$r[2]*$f);
	    push @{$self->{'_data'}},$result;
            $self->{'_last_error'} = '8'.$r[3];
	    $self->{'_last_warning'} = $r[4];
	    if ($r[4] =~ /(\d+)/)
	    {
		$self->{'_last_warning'} = $1;
	    }
	    return $result;
	}
	sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
	next OUTER if (!$testfile || -e $testfile);
        $self->reset();
	push @{$self->{'_data'}},undef;
        $self->{'_last_error'} = 'XX';
        return undef;
    }    
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}

sub run_frequency_all($$)
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    $result .= ' #';
    $self->writecmd('SO?');
    my $r = $self->readcmd();
    my $org = undef;
    if ($r =~ /([\d\,]+)/)
    {
        $org = 'SO '.$1;
    }
    foreach my $m (@{$self->{'_values_source_mode'}})
    {
        $self->checkcmd('SO '.$self->{'_source_modes'}{$m});
	$self->writecmd('DO'); ## write all to gpib ##
        my $res = $self->readcmd();
	my $count = 5;
	while (!$res && $count--)
	{
	    ### Try again a few times ####
	    $self->checkcmd('SO '.$self->{'_source_modes'}{$m});
	    $self->writecmd('DO'); ## write all to gpib ##
	    $res = $self->readcmd();   
	}
	### Supplpy dummy data if 5 retries didn't help ###
	$res = '0.0000,0.0000,0.0000' unless ($res);
        my @r = split(",",$res);
        $result .= sprintf(" $m %f %f %f",$r[0],$r[1],$r[2]);
    }
    $result .= ' #';
    foreach my $c (1,2)
    {
	$self->writecmd('RA'.$c.'?');
	my $res = $self->readcmd();
	unless (defined($res))
	{
	    ### Try again if first try did not wirk ###
	    $self->writecmd('RA'.$c.'?');
	    $res = $self->readcmd();	    
	}
        if (defined($res) && $res =~ /(\d+)/)
        {
            $res = $1+0;
        }
        if ($res)
        {
	    $result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
        }
        else
        {
            $self->writecmd('AR'.($c-1).'?');
            my $res = $self->readcmd();
            if (defined($res) && $res =~ /(\d+)/)
            {
                $res = $1+0;
            }
	    $result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
        }
	$self->writecmd('ER?');
	$self->readcmd();
    }
    $self->checkcmd($org);
    $self->writecmd('DO'); ## write all to gpib ##
    push @{$self->{'_longdata'}},$result;
    return $result;
}



package Solartron1255B;
our @ISA = qw(Solartron1255);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1255B';
    $self->{'_maxf'} = 1e6;
    bless $self,$class;
    return $self;
}

package Solartron1252;
our @ISA = qw(Solartron1255);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1252';
    $self->{'_maxf'} = 3e5;
    bless $self,$class;
    return $self;
}

return 1;
