package Solartron1260;
#$Id: Solartron1260.pm,v 1.69 2017/10/26 14:16:32 sofc Exp $

############################################################
#
#    Package Solartron1260
#
# This module implements the FRA class for a Solartron
# --{\textregistered} 
# 1260 frequency response analyser.
#
# To get a device instance call the constructor:
#
# $fra = Solartron1260->new($gpib_address);
#
############################################################

use strict;
use warnings;
use FDEV;
our @ISA = qw(FRA);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1260';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_minf'} = 10E-6;
    $self->{'_maxf'} = 32E+6;
    $self->{'_minpV'} = -40.95;
    $self->{'_maxpV'} = 40.95;
    $self->{'_minpA'} = -0.1;
    $self->{'_maxpA'} = 0.1;
    $self->{'_maxampA'} = 0.06;
    $self->{'_maxampV'} = 3;
    $self->{'_source_modes'} = { 'V1' => '1,0' , 'V2' => '2,0' , 'I' => '3,0', 'V1/I' => '1,3' , 'V2/I' => '2,3' , 'V1/V2' => '1,2' , 'V2/V1' => '2,1'};
    $self->{'_warningcodes'} = {'-1' => 'Low limit', '00' => ' ', '+1' => 'High limit'};
    $self->{'_errorcodes'} = {
        '00' => ' ',
        '01' => 'Unknown command',
        '02' => 'Argument mismatch',
        '03' => 'Argument out of range',
        '04' => 'Floating point format error',
        '05' => 'Ilegal request for value',
        '06' => 'Invalid function',
        '07' => 'Number out of range',
        '08' => 'Invalid symbol',
        '09' => 'Amplitude illegal for f > 10MHz',
        '31' => 'Generator killed',
        '32' => 'Generator overload',
        '34' => 'Generator restart',
	'XX' => 'No data recieved',
        '80' => ' ',
        '81' => 'Input Overload',
        '82' => 'Auto integration failed',
        '83' => 'Input Overload and Autointegration failed',
        '84' => 'Monitor failed',
        '85' => 'Input Overload and Monitor fail',
        '86' => 'Autointegration failed and Monitor fail',
        '87' => 'Input overload, Monitor fail and Autointegration failed',
        '88' => 'Auto impedance error',
        '89' => 'Group dealy parameter not set',
    };
    $self->{'_overhead'} = 1.05;
    $self->{'_minintegration'} = 0.01;
    $self->{'_setuptime'} = 6;
    $self->{'_delay'} = 0.1;  ## Default delay of 100 ms ##
    $self->{'_setup_tags'} = 'mode,input_mode_V1,input_mode_V2,monitor_enable,monitor_voltage_limit,monitor_current_limit,monitor_error,polarisation,polarisation_offset,source_mode,outer_mode_V1,outer_mode_V2,input_coupling_V1,input_coupling_V2,input_coupling_I,measuring_range_V1,measuring_range_V2,measuring_range_I,overhead,amplitude,delay';
    $self->{'_default'} = {
	'mode' => 'Volt',
	'input_mode_V1' => 'Differential',
	'input_mode_V2' => 'Differential',
	'source_mode' => 'V1/I',
	'outer_mode_V1' => 'Grounded',
	'outer_mode_V2' => 'Grounded',
	'input_coupling_V1' => 'DC',
	'input_coupling_V2' => 'DC',
	'input_coupling_I' => 'DC',
	'measuring_range_V1' => 'auto',
	'measuring_range_V2' => 'auto',
	'measuring_range_I' => 'auto',
	'monitor_enable' => 'NO',
	'monitor_error' => 50,
	'polarisation_offset' => 'OFF',
	'reference_potential' => 'Fixed',
        'delay' => $self->{'_delay'},
        'amplitude' => $self->{'_minampV'},
        'polarisation' => 0,
    };
    $self->{'_desc'} = {
	'mode' => 'Control mode',
	'input_mode_V1' => 'Input mode for V1 input, either differential or single ended',
	'input_mode_V2' => 'Input mode for V2 input, either differential or single ended',
	'source_mode' => 'Determines how the impedance is calculated',
	'outer_mode_V1' => 'Determines if the screen on voltage input 1 is grounded or left floating',
	'outer_mode_V2' => 'Determines if the screen on voltage input 2 is grounded or left floating',
	'input_coupling_V1' => 'Determines if the input coupling for voltage input 1 is direct through or uses a high pass filter',
	'input_coupling_V2' => 'Determines if the input coupling for voltage input 2 is direct through or uses a high pass filter',
	'input_coupling_I' => 'Determines if the input coupling for the current sense is direct through or uses a high pass filter',
	'measuring_range_V1' => 'Input range for voltage input 1, either fixed or autorange',
	'measuring_range_V2' => 'Input range for voltage input 2, either fixed or autorange',
	'measuring_range_I' => 'Input range for current input, either fixed or autorange',
	'monitor_enable' => 'Determines if input monitoring is to be used (device tries to addjust output so input matches given value)',
	'monitor_voltage_limit' => 'Maximum voltage if voltage limit is used (Volt)',
	'monitor_current_limit' => 'maximum current if current limit is used (0 to 6e-3 A)',
	'monitor_error' => 'maximum error in percent (between 1 and 50) when monitoring is used (failure to do this gives an error)',
	'polarisation_offset_program' => 'Select if a polarisation offset program (and which one) is to be run prior to impedance scans',
	'reference_potential' => 'Select if the polarisation is to be set versus chassis ground or versus measured potential at OCV',
        'delay' => 'Delay in seconds between each frequency',
        'amplitude' => 'AC amplitude (RMS, V or mA depending on mode)',
        'polarisation' => 'DC polarisation (V or mA depending on mode)',
    };
    $self->{'_values_source_mode'} = ['V1' , 'V2' , 'I', 'V1/I' , 'V2/I'  , 'V1/V2', 'V2/V1'];
    $self->{'_values_outer_mode_V1'} = ['Floating','Grounded'];
    $self->{'_values_outer_mode_V2'} = ['Floating','Grounded'];
    $self->{'_values_input_coupling_V1'} = ['AC','DC'];
    $self->{'_values_input_coupling_V2'} = ['AC','DC'];
    $self->{'_values_input_coupling_I'} = ['AC','DC'];
    $self->{'_values_measuring_range_V1'} = ['auto',0.03,0.3,3];
    $self->{'_values_measuring_range_V2'} = ['auto',0.03,0.3,3];
    $self->{'_values_measuring_range_I'} = ['auto',6e-6,6e-5,6e-4,6e-3,6e-2];
    $self->{'_values_input_mode_V1'} = ['Differential','Single'];
    $self->{'_values_input_mode_V2'} = ['Differential','Single'];
    $self->{'_values_monitor_enable'} = ['Current','Potential','NO'];
    $self->{'_save_tags'} = ['GT','FR','VA','VB','IA','IB','ME','VC','IC','AE','MS','RA1','RA2','RA3','DC3','DC1','DC2','IP1','IP2','OU1','OU2','SO','CV','VI','CI','FV','CZ','CY'];
    bless $self,$class;
    return $self;
}


sub bulk_load_check(@)
{
    my $self = shift;
    $self->writecmd('CE');
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OP2,1');  # GPIB does all #
    while (@_)
    {
        my $cmd = shift;
        $self->checkcmd($cmd);
    }
    return undef;    
}

sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
	push @res,$tag.' '.$self->readvalue($tag);
    }
    return @res;
}

sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    $self->writecmd('*RST');
    sleep(3);
    $self->writecmd('BP0');
    $self->writecmd('BP0');
    $self->writecmd('CE');
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OP2,1');  # GPIB does all #
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    $self->writecmd('VA 0.000');
    $self->writecmd('VB 0.000');
    $self->writecmd('IA 0.000');
    $self->writecmd('IB 0.000');
    return $res;
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OP2,1');  # GPIB does all #
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /solartron/i);
    return 0;
}

sub reset()
{
    my $self = shift;
    $self->writecmd('BK');
    sleep(1);
    $self->writecmd('BK');
    sleep(1);
    $self->writecmd('OT3'); # Fast reply cr + EOI #
    sleep(1);
    $self->writecmd('OP2,1');  # GPIB does all #
}

sub set_polarisation($$)
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
    if (ref($ext_pol) eq 'RemoteExec')
    {
	$self->writedebug("Executeing function to get polarisation offset");
        $self->{'_pol_offset'} = $ext_pol->exec_id($self);
        $pol += $self->{'_pol_offset'};
    }
    if ($self->{'_mode'} =~ /A/i)
    {
	$pol = $self->{'_minpA'} if ($pol < $self->{'_minpA'});
	$pol = $self->{'_maxpA'} if ($pol > $self->{'_maxpA'});
	$self->checkcmd('GT 1');
	$self->checkcmd(sprintf("IB %.4g",$pol));
    }
    else
    {
	$pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
	$pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'});
	$self->checkcmd('GT 0');
	$self->checkcmd(sprintf("VB %.4g",$pol));
    }
    return $pol;
}

sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### runs command twice if first go results in error ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    $self->writecmd('ER?');
    my $e = $self->readcmd();
    if (!$e || $e !~ /^00/)
    {
        if ($e = /~(\d+)/)
        {
            $self->writedebug('Error description: '.$self->errorcode($1));
        }
        $self->writecmd('CE');
        $self->writecmd($cmd);
        $self->checkerror();
    }
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('ER?');
    my $rep = $self->readcmd();
    if (!$rep || $rep !~ /^00/)
    {
	$self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '$rep'");
        if ($rep = /~(\d+)/)
        {
            $self->errorlog('Error description: '.$self->errorcode($1));
        }
	$self->writecmd('CE');
    }
}

sub get_error()
{
    my $self = shift;
    $self->writecmd('ER?');
    my $rep = $self->readcmd();
    if ($rep =~ /(\d+)/)
    {
        my $r = $self->errorcode($1);
        $self->writedebug("Error description: $r");
        return $r;
    }
    return '';
}


sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    my $diff1 = 1;
    my $diff2 = 1;
    my $input = 1;

    $diff1 = 0 if ($self->get_setup_value($self->get_default('input_mode_V1'),$setup,'input_mode_V1') =~ /S/);
    $diff2 = 0 if ($self->get_setup_value($self->get_default('input:mode_V2'),$setup,'input_mode_V2') =~ /S/);
    $self->checkcmd("IP1,$diff1");
    $self->checkcmd("IP2,$diff2");
    $self->checkcmd('OT3'); # Fast reply: cr + EOI #
    $self->checkcmd('OP2,1');  # GPIB does all #
    my $monitor = 0;
    my $mstr = $self->get_setup_value($self->get_default('monitor_enable'),$setup,'monitor_enable');
    $monitor = 1 if ($mstr =~ /Volt/i);
    $monitor = 2 if ($mstr =~ /Curr/i);
    my $mV = $self->get_setup_value(0,$setup,'monitor_voltage_limit');
    my $mA = $self->get_setup_value(0,$setup,'monitor_current_limit');
    my $me = $self->get_setup_value($self->get_default('monitor_error'),$setup,'monitor_error');
    $self->checkcmd('ME '.$monitor);
    if ($me >=1 && $me <= 50)
    {
	$self->checkcmd('AE '.$me);
    }
    $self->checkcmd('WF0');  ## Sine wave ##
    $self->checkcmd('AU0');  ## Autointegration off ##
    $self->checkcmd('MD0');  ## Normal mode impedance ##
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    $self->checkcmd(sprintf("MS %.4g",$self->{'_delay'})); 
    $self->checkcmd('CZ0');  ## Display R,X ##
    my $source_mode = $self->get_setup_value($self->get_default('source_mode'),$setup,'source_mode');
    ### Default V1/I ###
    my $sm = '1,3';
    $sm = $self->{'_source_modes'}{$source_mode} if exists($self->{'_source_modes'}{$source_mode});
    ### Default V1/I ###
    $self->checkcmd('SO'.$sm);
    $self->checkcmd('CV0');  ## Display a,b ##
    $self->checkcmd('VI0'); ## Display frequency ##
    ### Screen to ground default on input ###
    my $ou1 = 0;
    $ou1 = 1 if ($self->get_setup_value($self->get_default('outer_mode_V1'),$setup,'outer_mode_V1') =~ /F/i);
    my $ou2 = 0;
    $ou2 = 1 if ($self->get_setup_value($self->get_default('outer:mode_V2'),$setup,'outer_mode_V2') =~ /F/i);
    $self->checkcmd('OU1,'.$ou1);
    $self->checkcmd('OU2,'.$ou2);
    ### DC coupling default ###
    my $ac1 = 0;
    $ac1 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V1'),$setup,'input_coupling_V1') =~ /AC/i);
    my $ac2 = 0;
    $ac2 = 1 if ($self->get_setup_value($self->get_default('input_coupling_V2'),$setup,'input_coupling_V2') =~ /AC/i);
    my $ac3 = 0;
    $ac3 = 1 if ($self->get_setup_value($self->get_default('input_coupling_I'),$setup,'input_coupling_I') =~ /AC/i);
    $self->checkcmd('DC1,'.$ac1);
    $self->checkcmd('DC2,'.$ac2);
    $self->checkcmd('DC3,'.$ac3);
    ### Measure range on input ###
    my $ra1 = $self->get_setup_value($self->get_default('measuring_range_V1'),$setup,'measuring_range_V1');
    if ($ra1)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V1'}})
        {
            $tr = $x if ($_ eq $ra1);
            $x++;
	}
        $ra1 = $tr;
    }
    my $ra2 = $self->get_setup_value($self->get_default('measuring_range_V2'),$setup,'measuring_range_V2');
    if ($ra2)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_V1'}})
        {
            $tr = $x if ($_ eq $ra2);
            $x++;
	}
	$ra2 = $tr;
    }
    $self->checkcmd('RA1,'.$ra1);
    $self->checkcmd('RA2,'.$ra2);
    my $rai = $self->get_setup_value($self->get_default('measuring_range_I'),$setup,'measuring_range_I');
    if ($rai)
    {
        my $tr = 0; ## Default autorange ##
        my $x = 0;
        foreach (@{$self->{'_values_measuring_range_I'}})
        {
            $tr = $x if ($_ eq $ra1);
            $x++;
	}
	$rai = $tr;
    }
    $self->checkcmd('RA3,'.$ra2);
    $self->checkcmd('SW0'); ## Sweep off (we do all)
    ### Set polarisation and amplitude ###
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol,$setup);
    $self->amplitude($self->get_setup_value(0,$setup,'amplitude'));
    $self->reset_data;
    return $truepol;
}

sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $amp = shift;
    if ($self->{'_mode'} =~ /A/i)
    {
	$amp = $self->{'_minampA'} if ($amp < $self->{'_minampA'});
	$amp = $self->{'_maxampA'} if ($amp > $self->{'_maxampA'}); 
	$self->checkcmd(sprintf("IA %.4g",$amp));  ## Current AC amplitude ##
    }
    else
    {
	$amp = $self->{'_minampV'} if ($amp < $self->{'_minampV'});
	$amp = $self->{'_maxampV'} if ($amp > $self->{'_maxampV'});
	$self->checkcmd(sprintf("VA %.4g",$amp));  ## Voltage AC amplitude ##
    }
    $self->{'_amplitude'} = $amp;
    return $self->{'_amplitude'};
}


sub readvalue()
{
    my $self = shift;
    my $cmd = shift;
    my $res = undef;
    my $count = 0;
    while (!$res && $count++ < 5)
    {
	$self->writecmd($cmd.'?');
	my $r = $self->readcmd();
	if ($r =~ /([\d\,\.\+\-eE]+)/)
	{
	    $res = $1;
	}
    }
    return $res if (defined($res));
    $count--;
    my $str = "Could not read value for tag '$cmd', tried $count times!";
    $self->errorlog($str);
    return undef;
}

sub run_frequency_all($$)
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    $result .= ' #';
    my $org = 'SO '.$self->readvalue('SO');
    foreach my $m (@{$self->{'_values_source_mode'}})
    {
	$self->checkcmd('SO '.$self->{'_source_modes'}{$m});
	$self->writecmd('DO');
	my $res = $self->readcmd();
	my $count = 5;
	while (!$res && $count--)
	{
	    ### Try once more if first try did not work ###
	    $self->checkcmd('SO '.$self->{'_source_modes'}{$m});
	    $self->writecmd('DO');
	    $res = $self->readcmd();    
	}
	#### Supply dummydata if 5 retries didn't work ###
	$res = '0.000,0.000,0.000' unless ($res);
	my @r = split(",",$res);
	$result .= sprintf(" $m %f %f %f",$r[0],$r[1],$r[2]);
    }
    $result .= ' #';
    foreach my $c (1,2,3)
    {
	my  $res = $self->readvalue('RA'.$c);
	if (defined($res) && $res =~ /(\d+)/)
	{
	    $res = $1+0;
	}
	if ($res)
	{
	    if ($c < 3)
	    {
		$result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
	    }
	    else
	    {
		$result .= " Range I: ".$self->{'_values_measuring_range_I'}[$res];
	    }
	}
	else
	{
	    my  $res = $self->readvalue('AR'.($c-1));
	    if (defined($res) && $res =~ /(\d+)/)
	    {
		$res = $1+0;
	    }
	    if ($c < 3)
	    {
		$result .= " Range V$c: ".$self->{'_values_measuring_range_V'.$c}[$res];
	    }
	    else
	    {
		$result .= " Range I: ".$self->{'_values_measuring_range_I'}[$res];
	    }	    
	}
    }
    $self->checkcmd($org) if ($org =~ /\,/);
    $self->writecmd('DO');
    $self->readcmd(); ## To make sure output is ok, unknown why ###
    push @{$self->{'_longdata'}},$result;
    return $result;
}


sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});    
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    $self->checkcmd('FR '.$fout);
    $self->checkcmd('IS '.$tout);
    $self->writecmd('SI');  ## Run impedance ###
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
	sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
	next if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
	my $readloop = 0;
      INNER:
	while ($readloop < 5)
	{
	    $self->writecmd('DO');
	    my $res = $self->readcmd();
	    if (!$res || $res =~ /error/i)
	    {
		$readloop++;
		next INNER;
	    }
	    my @r = split(",",$res);
		# sometimes the reading is an response of the old recording, in that case a new reading of the solartron buffer give the new recording result
		# !note if single frequency measurement is done this will not trap old data reading from the past measurement
		if($r[0] != $fout){
				$self->writecmd('DO'); ## write all to gpib ##
				$res = $self->readcmd();	## perhaps check for error again
				@r = split(",",$res);
		}
	    my $string = sprintf("%f %f %f",$r[0],$r[1]*$f,$r[2]*$f);
	    push @{$self->{'_data'}},$string;
	    $self->{'_last_error'} = '8'.$r[3];
	    $self->{'_last_warning'} = $r[4];
            if ($r[4] =~ /(\d+)/)
            {
                $self->{'_last_warning'} = $1;
            }
	    return $string;
	}
	sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
	next OUTER if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	$self->{'_last_error'} = 'XX';
	return undef;
    }
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}

return 1;
