package VirtualFRA;
#$Id: VirtualFRA.pm,v 1.11 2018/10/30 07:24:56 sofc Exp $

#####################################################################
#
#
#   Package VirtualFRA
#
#
#   This module implementes a virtual frequency response analyser.
#   
#   To get a device instance call the constructor:
#   
#   $fra = VirtualFRA->new($gpib_address);
#
#   Note that the gpib-address is only included for portability, it is
#   not used in any way by the virtual device.
#
#   The actual impednace values returnded is determined by the 'command'
#   setting which should be a valid executable which MUST be placed
#   in the directory
#   ----\\ 
#   /usr/local/bin/elchemea/user_exec/
#
#   The program will be called with the following arguments:
#
#   frequency, integration_time
#
#   And must return the impedance in the form of a string:
#
#   Freq Real Imag
#
#####################################################################


use strict;
use warnings;
use FDEV;
use Math::Complex;
our @ISA = qw(FRA);


sub new
{
    my $class = shift;
    my $addr = shift;
    $addr = 0;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Virtual FRA';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V'};
    $self->{'_minf'} = 10E-6;
    $self->{'_maxf'} = 32E+6;
    $self->{'_minpV'} = -40.95;
    $self->{'_maxpV'} = 40.95;
    $self->{'_minpA'} = -0.1;
    $self->{'_maxpA'} = 0.1;
    $self->{'_maxampA'} = 0.06;
    $self->{'_maxampV'} = 3;
    $self->{'_overhead'} = 1.05;
    $self->{'_minintegration'} = 1;
    $self->{'_setuptime'} = 6;
    $self->{'_delay'} = 0.1;  ## Default delay of 100 ms ##
    $self->{'_setup_tags'} = 'mode,overhead,amplitude,polarisation,delay,command,polarisation_offset_program,polarisation_offset';
    $self->{'_default'} = {
        'mode' => 'Volt',
        'source_mode' => 'V2/V1',
        'polarisation_offset' => 'OFF',
        'reference_potential' => 'Fixed',
        'delay' => $self->{'_delay'},
        'amplitude' => $self->{'_minampV'},
        'polarisation' => 0
    };
    $self->{'_desc'} = {
        'mode' => 'Control mode',
        'polarisation_offset_program' => 'Select if a polarisation offset program (and which one) is to be run prior to impedance scans',
        'delay' => 'Delay in seconds between each frequency',
        'amplitude' => 'AC amplitude (RMS, V)',
        'polarisation' => 'DC polarisation (V)',
        'source_mode' => 'Determines how the impedance is calculated',
	'command' => 'Command to execute to get impedance value'
    };
    $self->{'_values_source_mode'} = ['V2/V1'];
    bless $self,$class;
    return $self;
}

sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    return 1;
}

sub reset()
{
    my $self = shift;
    sleep(1);
}

sub isallive()
{
    my $self = shift;
    return 1;
}

sub set_polarisation($$)
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
    if (ref($ext_pol) eq 'RemoteExec')
    {
        $self->writedebug("Executeing function to get polarisation offset");
        $self->{'_pol_offset'} = $ext_pol->exec_id($self);
        $pol += $self->{'_pol_offset'};
    }
    return $pol;
}

sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    ### Set polarisation and amplitude ###
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol,$setup);
    $self->amplitude($self->get_setup_value(0,$setup,'amplitude'));
    $self->reset_data;
    return $truepol;
}

sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $a = shift;
    $a = $self->{'_minampV'} if ($a < $self->{'_minampV'});
    $a = $self->{'_maxampV'} if ($a > $self->{'_maxampV'});
    $self->{'_amplitude'} = $a;
    return $self->{'_amplitude'};
}


sub run_frequency_all($$)
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    push @{$self->{'_longdata'}},$result;
    return $result;
}

sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});    
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    my $tout = sprintf("%.4g",$time);
    my $cmd = $$setup{'command'};
    chomp $cmd if ($cmd);
    my $res;
    if ($cmd && $cmd =~ /^([\w\s\-\.]+)$/)
    {
	$cmd = '/usr/local/bin/elchemea/user_exec/'.$1." $freq $time ".$self->{'_pol_offset'};
	$self->writedebug($cmd);
	$res = `$cmd`;
	$self->writedebug($res);
	$res =~ s/[\r\n]+//g;
    }
    else
    {
	$res = sprintf("%f %f %f",$freq,1,0);
    }
    ## Emulate errors in transmission, time can be up to 20% longer ##
    $time = $time * (1 + (rand(2)/10));
    my $etime = time + $time;
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next if (!$testfile || -e $testfile);
        $self->reset();
        push @{$self->{'_data'}},undef;
        return undef;
    }
    push @{$self->{'_data'}},$res;
    return $res;
}

return 1;
